/*
 * Copyright (C) 2006 Changbeom Ahn
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "pz.h"
#include "tihangul.h"


/* dependent on textinput module */
extern ttk_color ti_ap_get(int);
extern TApItem * ti_ap_getx(int);
extern int ti_register(TWidget *(* cr)(), TWidget *(* ncr)(), char *, int);
extern TWidget * ti_create_tim_widget(int ht, int wd);


static PzModule * module;


#define TI_HANGUL_TIM_ID	28

#define TI_HANGUL_NO_INDEX	0x7F
#define TI_HANGUL_NUM_BUTTONS	4
#define TI_HANGUL_PREVIEW_AREA_WIDTH	20


#define TI_HANGUL_NEXT(n) \
	(ti_hangul_state_tab[ti_hangul_state].next##n)


static ti_hangul_index_t ti_hangul_initial	= TI_HANGUL_NO_INDEX;
static ti_hangul_index_t ti_hangul_peak		= TI_HANGUL_NO_INDEX;
static ti_hangul_index_t ti_hangul_final;
static char ti_hangul_state;


static uc16
ti_hangul_char()
{
	/*
	printf("ti_hangul_char: i=%d p=%d f=%d\n",
			ti_hangul_initial, ti_hangul_peak, ti_hangul_final);
	*/

	if (ti_hangul_initial == TI_HANGUL_NO_INDEX) {
		return 0;
	}

	if (ti_hangul_peak == TI_HANGUL_NO_INDEX) {
		return ti_hangul_initial_tab[ti_hangul_initial];
	}

	uc16 off = ti_hangul_initial * TI_HANGUL_NUM_PEAKS;
	off = (off + ti_hangul_peak) * TI_HANGUL_NUM_FINALS + ti_hangul_final;
	return 0xAC00 + off;
}


static uc16 const*
ti_hangul_str()
{
	/*
	printf("ti_hangul_str: enter\n");
	*/

	static uc16 str[2];

	str[0] = ti_hangul_char();
	/*
	printf("ti_hangul_str: char=%X\n", str[0]);

	printf("ti_hangul_str: leave\n");
	*/
	return str;
}


ti_hangul_index_t
ti_hangul_next(int n)
{
	printf("thn: enter\n");
	if (n == 1) {
		return TI_HANGUL_NEXT(1);
	} else if (n == 2) {
		return TI_HANGUL_NEXT(2);
	} else if (n == 3) {
		return TI_HANGUL_NEXT(3);
	} else if (n == 4) {
		return TI_HANGUL_NEXT(4);
	} else {
		fprintf(stderr, "ti_hangul_next: invalid n: %d\n", n);
	}

	return 0;
}


uc16 const*
ti_hangul_button_text(int n)
{
	static uc16 result[2] = {0, };
	ti_hangul_index_t state = ti_hangul_next(n);
	printf("button_text: n=%d, state=%d\n", n, state);
	/*
	printf("ti_hangul_button_text: enter\n");
	*/

	if (TI_HANGUL_IS_INITIAL(state)) {
		result[0] = ti_hangul_initial_tab[
			TI_HANGUL_INITIAL_INDEX(state)];
	} else if (TI_HANGUL_IS_PEAK(state)) {
		if (n == 1) {
			result[0] = ti_hangul_peak_tab[
				TI_HANGUL_PEAK_INDEX(TI_HANGUL_STATE_VBAR)];
		} else if (n == 2) {
			return TI_HANGUL_TEXT(TI_HANGUL_STATE_DOT);
		} else if (n == 3) {
			result[0] = ti_hangul_peak_tab[
				TI_HANGUL_PEAK_INDEX(TI_HANGUL_STATE_HBAR)];
		} else {
			fprintf(stderr, "ti_hangul_button_text: unexpected n "
					"for peak: %d", n);
		}
	} else if (TI_HANGUL_IS_FINAL(state)) {
		result[0] = ti_hangul_final_tab[
			TI_HANGUL_FINAL_INDEX(state)];
	} else {
		/*
		printf("ti_hangul_button_text: state=%d\n", state);
		printf("ti_hangul_button_text: leave\n");
		*/
		if (state == TI_HANGUL_STATE_DOTDOT) {
			state = TI_HANGUL_STATE_DOT;
		}
		return TI_HANGUL_TEXT(state);
	}

	/*
	printf("ti_hangul_button_text: leave\n");
	*/
	return result;
}


static void
ti_hangul_text_uc16(ttk_surface srf, int x, int y, uc16 const* str)
{
	printf("ti_hangul_text_uc16: enter\n");
	/*
	printf("ti_hangul_text_uc16: str=%p, *str=%u\n", str, *str);
	*/
	ttk_text_uc16(srf, ttk_textfont, x, y, ti_ap_get(1), str);
	/*
	printf("ti_hangul_text_uc16: leave\n");
	*/
}


static void
ti_hangul_draw(TWidget * wid, ttk_surface srf)
{
	int i;

	/*
	printf("ti_hangul_draw: enter\n");
	*/

	ti_hangul_text_uc16(srf, wid->x, wid->y, ti_hangul_str());

	for (i = 0; i < TI_HANGUL_NUM_BUTTONS; ++i) {
		printf("thd: i=%d\n", i);
		int x = TI_HANGUL_PREVIEW_AREA_WIDTH +
			i * (wid->w - TI_HANGUL_PREVIEW_AREA_WIDTH) / 4;
		ti_hangul_text_uc16(srf, wid->x + x, wid->y,
				ti_hangul_button_text(i + 1));
	}

	/*
	printf("ti_hangul_draw: leave\n");
	*/
}


static int
ti_hangul_scroll(TWidget * wid, int dir)
{
	/*
	TTK_SCROLLMOD (dir, 4);
	
	if (ti_hangul_numeric) {
		if (dir<0) {
			ti_hangul_pos = ((ti_hangul_pos + 16) % 17);
		} else {
			ti_hangul_pos = ((ti_hangul_pos + 1) % 17);
		}
	} else {
		if (dir<0) {
			ti_hangul_digit[ti_hangul_pos] = ((ti_hangul_digit[ti_hangul_pos]-1) & 0xF);
		} else {
			ti_hangul_digit[ti_hangul_pos] = ((ti_hangul_digit[ti_hangul_pos]+1) & 0xF);
		}
	}
	wid->dirty = 1;
	*/
	return TTK_EV_CLICK;
}


static void
ti_hangul_update(TWidget* wid)
{
	if (TI_HANGUL_IS_INITIAL(ti_hangul_state)) {
		ti_hangul_initial = TI_HANGUL_INITIAL_INDEX(ti_hangul_state);
	} else if (TI_HANGUL_IS_PEAK(ti_hangul_state)) {
		ti_hangul_peak = TI_HANGUL_PEAK_INDEX(ti_hangul_state);
	} else if (TI_HANGUL_IS_FINAL(ti_hangul_state)) {
		ti_hangul_final = TI_HANGUL_FINAL_INDEX(ti_hangul_state);
	} else {
		printf("ti_hangul_update: no group\n");
		/* TODO: handle exception */
	}

	wid->dirty = 1;
}


#define TI_HANGUL_HANDLE_BUTTON(n) \
	do { \
		if (TI_HANGUL_NEXT(n) == TI_HANGUL_STATE_XX) break; \
		TI_HANGUL_SET_STATE(TI_HANGUL_NEXT(n)); \
		printf("state=%d\n", ti_hangul_state); \
		ti_hangul_update(wid); \
	} while (0)


#define TI_HANGUL_SET_STATE(next) \
	ti_hangul_state = (next)


static void
ttk_hangul_reset()
{
	ti_hangul_initial = ti_hangul_peak = TI_HANGUL_NO_INDEX;
	ti_hangul_final = 0;
	ti_hangul_state = 0;
}


static int
ti_hangul_down(TWidget * wid, int btn)
{
	/*
	printf("ti_hangul_down: enter\n");
	*/

	switch (btn) {
	case TTK_BUTTON_PLAY:
		TI_HANGUL_HANDLE_BUTTON(3);
		break;
	case TTK_BUTTON_ACTION:
		{
			uc16 const code = ti_hangul_char();
			if (code) {
				ttk_input_char(code);
				ttk_hangul_reset();
				wid->dirty = 1;
			}
		}
		break;
	case TTK_BUTTON_PREVIOUS:
		TI_HANGUL_HANDLE_BUTTON(1);
		break;
	case TTK_BUTTON_NEXT:
		TI_HANGUL_HANDLE_BUTTON(4);
		break;
	case TTK_BUTTON_MENU:
		TI_HANGUL_HANDLE_BUTTON(2);
		break;
	default:
		return TTK_EV_UNUSED;
		break;
	}

	/*
	printf("ti_hangul_down: leave\n");
	*/

	return TTK_EV_CLICK;
}


static int
ti_hangul_held(TWidget* wid, int btn)
{
	if (btn == TTK_BUTTON_MENU) {
		ttk_input_end();
	} else if (btn == TTK_BUTTON_PREVIOUS) {
		ttk_input_char('\b');
	} else if (btn == TTK_BUTTON_PLAY) {
		ttk_input_char(' ');
	} else if (btn == TTK_BUTTON_NEXT) {
		ttk_input_char('\n');
	} else {
		return TTK_EV_UNUSED;
	}
	ttk_hangul_reset();
	++wid->dirty;
	return TTK_EV_CLICK;
}


static TWidget*
ti_hangul_create(int n)
{
	TWidget * wid = ti_create_tim_widget(-1, 0);
	wid->draw = ti_hangul_draw;
	wid->down = ti_hangul_down;
	wid->scroll = ti_hangul_scroll;
	wid->held = ti_hangul_held;
	return wid;
}


static TWidget*
ti_hangul_screate()
{
	return ti_hangul_create(0);
}


static TWidget*
ti_hangul_ncreate()
{
	return ti_hangul_create(1);
}


static void
ti_hangul_init()
{
	module = pz_register_module("tihangul", NULL);
	ti_register(ti_hangul_screate, ti_hangul_ncreate, N_("Hangul"),
			TI_HANGUL_TIM_ID);
}


PZ_MOD_INIT(ti_hangul_init)
